import logging
import datetime
import io

import flask
import matplotlib.cm
import matplotlib.colors

from . import datasets
from . import utils
from . import plots

logger = logging.getLogger(__name__)


def index(api: object) -> str:
    # can't name this index (already taken by conexxion)
    return flask.render_template("main.html", api=api)


def transect(id) -> object:
    logger.info(flask.request)
    return {}


def transect_overview() -> list:
    logger.info(flask.request)
    return []


def transect_overview_kml() -> str:
    """create an overview of all transects"""
    lines = datasets.overview()
    return flask.render_template('lod.kml', lines=lines)


def transect_kml(
        id: int,
        extrude: bool,
        exaggeration: float,
        lift: float,
        move: float
) -> str:
    """create a kml for a specific transect"""
    # only available on runtime
    flask.current_app.jinja_env.filters['kmldate'] = utils.kmldate
    transect = datasets.get_transect(int(id), exaggeration, lift, move)
    return flask.render_template("transect.kml", transect=transect, extrude=int(extrude))


def transect_info(id: int) -> str:
    return flask.render_template("info.html")


def time_map(id: int) -> str:
    data = datasets.get_transect_data(int(id))
    fig, ax = plots.time_map(data)
    stream = io.BytesIO()
    fig.savefig(stream)
    return stream.getvalue()


def eeg(id: int) -> str:
    data = datasets.get_transect_data(int(id))
    fig, ax = plots.eeg(data)
    stream = io.BytesIO()
    fig.savefig(stream)
    return stream.getvalue()


def styles(poly_alpha: float, outline: int, colormap: str) -> str:
    """return style information"""

    context = {}

    poly_alpha = int(float(poly_alpha) * 255)
    context['poly_alpha'] = '{:02X}'.format(poly_alpha)
    context['outline'] = int(outline)

    # Get a colormap based on the ?colormap parameter
    colormap_name = colormap
    colormap = matplotlib.cm.cmap_d.get(colormap_name, matplotlib.cm.viridis)
    colors = {}
    current_year = datetime.datetime.now().year
    N = matplotlib.colors.Normalize(1964, current_year + 1)
    for year in range(1964, current_year + 1):
        # call with float 0..1 (or int 0 .. 255)
        r, g, b, alpha = colormap(N(year))
        # r and b reversed in the google, don't forget to add alpha
        color = matplotlib.colors.rgb2hex((b, g, r)).replace('#', '')
        colors['year{0}'.format(year)] = color
    context['colors'] = colors
    return flask.render_template("styles.kml", context=context)



# -*- coding: utf-8 -*-

"""Console script for coastviewer."""

import logging
import functools

import click
import connexion

from . import controllers

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def make_app():
    """create the main application"""
    app = connexion.App(__name__, specification_dir='./swagger/')
    title = """
    This is the coastal-viewer api.
    It provides services to acquire data of coasts around the world.
    """
    api = app.add_api(
        'swagger.yaml',
        arguments={
            'title': title
        }
    )
    logger.debug(api)

    # make sure the function keeps the __name__ 'index' and __docs__
    index = functools.wraps(controllers.index)(
        # already fill in the api parameter
        functools.partial(controllers.index, api=api)
    )
    app.add_url_rule('/', 'index', index)

    return app


@click.command()
@click.option('--debug/--no-debug', default=False, help='Start application in debugger mode.')
def main(debug, args=None):
    """Console script for coastviewer."""
    # configure logging
    logging.basicConfig()

    app = make_app()
    app.run(debug=debug)


if __name__ == "__main__":
    main()



import itertools
import logging
import pathlib

import netCDF4
import pandas
import numpy as np
import pyproj
import matplotlib.cm
import scipy.interpolate

from . import utils

logger = logging.getLogger(__name__)

# allow to use a local dataset
local = True


DATASETS = {
    'transect': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/jarkus/profiles/transect.nc',  # nopep8
    'BKL_TKL_TND': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/BKL_TKL_MKL/BKL_TKL_TND.nc',  # nopep8
    'DF': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/DuneFoot/DF.nc',  # nopep8
    'mkl': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/BKL_TKL_MKL/MKL.nc',  # nopep8
    'strandbreedte': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/strandbreedte/strandbreedte.nc',  # nopep8
    'strandlijnen': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/MHW_MLW/MHW_MLW.nc',  # nopep8
    'suppleties': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/suppleties/suppleties.nc',  # nopep8
    'faalkans': 'http://opendap.deltares.nl/thredds/dodsC/opendap/rijkswaterstaat/faalkans_PC-Ring/faalkans.nc'  # nopep8
}

if local:
    DATASETS['transect'] = str(pathlib.Path('data/transect.nc'))

# global variables
with netCDF4.Dataset(DATASETS['transect']) as ds:
    # keep these global, for faster indexing
    ids = ds.variables['id'][:]


def overview():
    """generate a lod overview"""
    # read relevant variables
    with netCDF4.Dataset(DATASETS['transect']) as ds:
        variables = {
            'id': {"var": 'id', "slice": np.s_[:]},
            'lat_0': {"var": 'lat', "slice": np.s_[:, 0]},
            'lat_1': {"var": 'lat', "slice": np.s_[:, -1]},
            'lon_0': {"var": 'lon', "slice": np.s_[:, 0]},
            'lon_1': {"var": 'lon', "slice": np.s_[:, -1]},
            'rsp_lon': {"var": 'rsp_lon', "slice": np.s_[:]},
            'rsp_lat': {"var": 'rsp_lat', "slice": np.s_[:]}
        }
        data = {}
        for var, props in variables.items():
            data[var] = ds.variables[props['var']][props['slice']]
    # put in a table
    df = pandas.DataFrame(data=data)
    delta = 0.002
    df['north'] = df['rsp_lat'] + delta
    df['south'] = df['rsp_lat'] - delta
    df['east'] = df['rsp_lon'] + delta
    df['west'] = df['rsp_lon'] - delta

    # transform for kml file
    def line_coords(record):
        return utils.textcoordinates(
            x0=record.lon_0,
            y0=record.lat_0,
            x1=record.lon_1,
            y1=record.lat_1
        )
    df['line_coords'] = df.apply(line_coords, axis=1)

    def point_coords(record):
        return utils.textcoordinates(
            x0=record.rsp_lat,
            y0=record.rsp_lon
        )
    df['point_coords'] = df.apply(point_coords, axis=1)

    def bbox(record):
        return {
            'north': record.north,
            'south': record.south,
            'east': record.east,
            'west': record.west
        }
    df['bbox'] = df.apply(bbox, axis=1)

    n_pixels = itertools.islice(
        itertools.cycle([64, 32, 64, 16, 64, 32, 64, 8]),
        len(df)
    )
    df['min_lod_pixels'] = list(n_pixels)
    return df


def move_by(lon, lat, distance):
    """
    Move the x,y coordinates by distance, perpendicular, assuming that they are lat,lon and that we can move in EPSG:28992
    >>> lon = array([4.0])
    >>> lat = array([51.0])
    >>> x,y = move_by(lat, lon, 1000)
    >>> x, y  # doctest:+ELLIPSIS
    (array([ 3.999...]), array([ 51.0089...]))
    """
    # project from wgs84 to rd, assuming x,y are lon, lat
    # compute the angle from the transect coordinates
    rd = pyproj.Proj('+init=epsg:28992')

    x, y = rd(lon, lat)

    dx = x[-1] - x[0]
    dy = y[-1] - y[0]

    # rotate by 90 degrees
    angle = np.arctan2(dy, dx) + np.pi * 0.5

    x += distance * np.cos(angle)
    y += distance * np.sin(angle)

    lon, lat = rd(x, y, inverse=True)

    return lon, lat


def get_transect(id_, exaggeration=1.0, lift=0.0, move=0.0):
    """lookup information for transect"""

    transect_idx = np.searchsorted(ids, id_)
    variables = {
        'lat': {"var": 'lat', "slice": np.s_[transect_idx, :]},
        'lon': {"var": 'lon', "slice": np.s_[transect_idx, :]},
        'z': {"var": 'altitude', "slice": np.s_[:, transect_idx, :]},
        "t": {"var": 'time', "slice": np.s_[:]}

    }
    data = {}
    with netCDF4.Dataset(DATASETS['transect']) as ds:
        for var, props in variables.items():
            data[var] = ds.variables[props['var']][props['slice']]
        time_units = ds.variables['time'].units

    # df = pandas.DataFrame(data=data)
    years = []
    for i, (t, row) in enumerate(zip(data['t'], data['z'])):
        item = {}

        lon, lat = move_by(data['lon'], data['lat'], distance=i*move)
        coords = pandas.DataFrame(data=dict(
            lon=lon,
            lat=lat,
            z=row * exaggeration + lift
        ))
        item['coordinates'] = coords.dropna()
        item['line_coordinates'] = utils.textcoordinates(
            x0=item['coordinates']['lon'],
            y0=item['coordinates']['lat'],
            z0=item['coordinates']['z']
        )
        date = netCDF4.num2date(t, time_units)
        item['properties'] = {
            "t": date,
        }
        item['properties']['year'] = date.year
        item['properties']['begin_date'] = date
        item['properties']['end_date'] = date.replace(year=date.year+1)
        years.append(item)
    years = pandas.DataFrame.from_records(years)
    transect = {
        "years": years,
        "id": id_
    }
    return transect

def get_transect_data(id_=7003900):
    transect_idx = np.searchsorted(ids, id_)
    variables = {
        'lat': {"var": 'lat', "slice": np.s_[transect_idx, :]},
        'lon': {"var": 'lon', "slice": np.s_[transect_idx, :]},
        'z': {"var": 'altitude', "slice": np.s_[:, transect_idx, :]},
        "t": {"var": 'time', "slice": np.s_[:]},
        "cross_shore": {"var": "cross_shore", "slice": np.s_[:]},
        'mean_high_water': {"var": 'mean_high_water', "slice": np.s_[transect_idx]},
        'mean_low_water': {"var": 'mean_low_water', "slice": np.s_[transect_idx]},


    }
    data = {}
    with netCDF4.Dataset(DATASETS['transect']) as ds:
        for var, props in variables.items():
            data[var] = ds.variables[props['var']][props['slice']]
        time_units = ds.variables['time'].units
    data['time'] = netCDF4.num2date(data['t'], time_units)
    data['filled_z'] = fill(data['z'])
    data['time_num'] = matplotlib.dates.date2num(data['time'])
    return data

def fill(z):
    """fill z by space and then time"""
    def fill_space(z):
        """fill space"""
        x = np.arange(z.shape[0])
        F = scipy.interpolate.interp1d(x[~z.mask], z[~z.mask], bounds_error=False)
        z_interp = F(x)
        z_filled = np.ma.masked_invalid(z_interp)
        return z_filled
    def fill_time(z):
        """fill time"""
        z_filled = np.ma.masked_all_like(z)
        for i in range(z.shape[1]):
            arr = z[:, i]
            # if there's no data, continue
            if arr.mask.all():
                continue
            # interpolate in time
            xp = np.arange(len(arr))
            z_filled[:, i] = scipy.interp(xp, xp[~arr.mask], arr[~arr.mask])
        return z_filled

    filled_z = np.ma.apply_along_axis(fill_space, 1, z)
    filled_z = fill_time(filled_z)
    return filled_z



import pathlib
import matplotlib.pyplot as plt
import matplotlib.cm
import numpy as np
import netCDF4
import cmocean
import colorcet
import scipy.interpolate


import coastviewer.extra_cm

def time_map(data):
    fig, ax = plt.subplots(figsize=(8, 3))
    pc = ax.pcolorfast(data['cross_shore'], data['time_num'], data['filled_z'], vmin=-20, vmax=20, cmap=coastviewer.extra_cm.GMT_drywet_r)
    cont = ax.contour(
        data['cross_shore'], 
        data['time_num'], 
        data['z'],
        levels=[data['mean_low_water'], data['mean_high_water'], 3],
        cmap='PuBu'
    )
    # legend
    cb = plt.colorbar(pc, ax=ax)

    # labels
    cb.set_label('Height to NAP [m]')
    ax.set_xlabel('Cross shore distance [m]')
    ax.set_ylabel('Measurement time [y]')

    # date format
    date_locator = matplotlib.dates.AutoDateLocator()
    date_formatter = matplotlib.dates.AutoDateFormatter(date_locator)
    ax.yaxis.set_major_formatter(date_formatter)
    return fig, ax

def eeg(data):
    t = data['time_num']
    x = data['cross_shore']
    # and data
    z = data['z']
    nrows, nsamples = z.shape

    # create a line for each timeseries
    segs = []
    ticklocs = []
    for i, row in enumerate(z):
        # add a line, scale it by the y axis each plot has a range of the elevation divided by 7.5 (~2 years up and down)
        pts = np.c_[
            x[~z[i, :].mask],
            z[i, ~z[i,:].mask].filled()*365.0/7.5
        ]

        segs.append(pts)

        ticklocs.append(t[i]) # use date for yloc
    # create an offset for each line
    offsets = np.zeros((nrows, 2), dtype=float)
    offsets[:, 1] = ticklocs
    # create the lines
    lines = matplotlib.collections.LineCollection(segs, offsets=offsets)
    # create a new figure
    fig, ax = plt.subplots(figsize=(10, 6))
    ax.add_collection(lines)
    # set the x axis
    ax.set_xlim(x.min(), x.max())
    # set the y axis (add a bit of room cause the wiggles go over a few years)
    # changed maximum correction days to 1000 from 730 (2 years), because
    # sometimes upper line was outside the y limits
    ax.set_ylim(t.min()-730, t.max()+1000)
    ax.set_xlabel('Cross shore distance [m]')
    ax.set_ylabel('Measurement time [y]')

    date_locator = matplotlib.dates.AutoDateLocator()
    date_formatter = matplotlib.dates.AutoDateFormatter(date_locator)
    ax.yaxis.set_major_formatter(date_formatter)
    return fig, ax



# by convention.
import io
import zipfile
import re

import numpy as np


def compress_kml(kml):
    """
    Returns compressed KMZ from the given KML string.

    >>> kml = "<kml>"
    >>> # returns a zip file containing a doc.xml
    >>> compress_kml(kml)[:2]
    'PK'
    """

    kmz = io.BytesIO()
    zf = zipfile.ZipFile(kmz, 'a', zipfile.ZIP_DEFLATED)
    # kmz file is a zip file that contains a doc.kml file
    zf.writestr('doc.kml', kml)
    zf.close()
    kmz.seek(0)
    return kmz.read()


def textcoordinates(x0, y0, z0=None, x1=None, y1=None, z1=None):
    """
    convert the coordinates to a string so they can be used by kml

    # Example usage:
    >>> x = np.array([1,1,1])
    >>> y = np.array([1,-2e10,3.0000001])
    >>> print(textcoordinates(x, y))
    1.0,1.0,0.0
    1.0,-20000000000.0,0.0
    1.0,3.0000001,0.0
    <BLANKLINE>
    >>> # Allow for coordinate pairs
    >>> x0 = x
    >>> y0 = y
    >>> x1 = x
    >>> y1 = y
    >>> print(textcoordinates(x0=x0, y0=y0, x1=x1, y1=y1))
    1.0,1.0,0.0 1.0,1.0,0.0
    1.0,-20000000000.0,0.0 1.0,-20000000000.0,0.0
    1.0,3.0000001,0.0 1.0,3.0000001,0.0
    <BLANKLINE>
    >>> x = np.array([1])
    >>> y = np.array([1])
    >>> z = np.array([np.nan])
    >>> print(textcoordinates(x, y, z))
    <BLANKLINE>

    """
    if z0 is None:
        z0 = np.zeros_like(x0)
    if x1 is None:
        coordinates = np.vstack([x0, y0, z0]).T
    else:
        if z1 is None:
            z1 = np.zeros(x1.shape)
        coordinates = np.vstack([x0, y0, z0, x1, y1, z1]).T
    # only write coordinates where none is nan
    filter = np.isnan(coordinates).any(1)
    # use cStringIO for performance
    output = io.BytesIO()
    # save coordinates to string buffer
    # I think this is faster than other methods
    # Use %s to get reduced string output
    np.savetxt(output, coordinates[~filter], delimiter=',', fmt='%s')
    coord_bytes = output.getvalue()
    coord_str = coord_bytes.decode()
    if x1 is not None:
        # replace all the 3rd , by a space...
        # TODO double check this regex
        coord_str = re.sub(r'(.*?,.*?,.*?),(.*)',  r'\1 \2', coord_str)
    return coord_str


def kmldate(date):
    """
    print date in kml format

    # Example usage
    >>> import datetime
    >>> date = datetime.datetime.fromtimestamp(0)
    >>> # this breaks if tzdata changed after 1970, pff
    >>> date = datetime.datetime(year=2000, month=1, day=1)
    >>> kmldate(date)
    '2000-01-01T00:00:00Z'

    """
    return date.strftime("%Y-%m-%dT%H:%M:%SZ")
